@@ -25,30 +25,23 @@ try:
     import pandas
 except ImportError:  # pragma: NO COVER
     pandas = None
-try:
-    import IPython
-    from IPython.utils import io
-    from IPython.testing import tools
-    from IPython.terminal import interactiveshell
-except ImportError:  # pragma: NO COVER
-    IPython = None
 
 from google.api_core import exceptions
 import google.auth.credentials
 
-try:
-    from google.cloud import bigquery_storage_v1
-except ImportError:  # pragma: NO COVER
-    bigquery_storage_v1 = None
 from google.cloud import bigquery
 from google.cloud.bigquery import job
 from google.cloud.bigquery import table
-from google.cloud.bigquery import magics
+from google.cloud.bigquery.magics import magics
 from tests.unit.helpers import make_connection
 from test_utils.imports import maybe_fail_import
 
 
-pytestmark = pytest.mark.skipif(IPython is None, reason="Requires `ipython`")
+IPython = pytest.importorskip("IPython")
+io = pytest.importorskip("IPython.utils.io")
+tools = pytest.importorskip("IPython.testing.tools")
+interactiveshell = pytest.importorskip("IPython.terminal.interactiveshell")
+bigquery_storage_v1 = pytest.importorskip("google.cloud.bigquery_storage_v1")
 
 
 @pytest.fixture(scope="session")
@@ -69,6 +62,21 @@ def ipython_interactive(request, ipython):
         yield ipython
 
 
+@pytest.fixture()
+def ipython_ns_cleanup():
+    """A helper to clean up user namespace after the test
+
+    for the duration of the test scope.
+    """
+    names_to_clean = []  # pairs (IPython_instance, name_to_clean)
+
+    yield names_to_clean
+
+    for ip, name in names_to_clean:
+        if name in ip.user_ns:
+            del ip.user_ns[name]
+
+
 @pytest.fixture(scope="session")
 def missing_bq_storage():
     """Provide a patcher that can make the bigquery storage import to fail."""
@@ -256,7 +264,7 @@ def test__run_query():
     ]
 
     client_patch = mock.patch(
-        "google.cloud.bigquery.magics.bigquery.Client", autospec=True
+        "google.cloud.bigquery.magics.magics.bigquery.Client", autospec=True
     )
     with client_patch as client_mock, io.capture_output() as captured:
         client_mock().query(sql).result.side_effect = responses
@@ -284,7 +292,7 @@ def test__run_query_dry_run_without_errors_is_silent():
     sql = "SELECT 17"
 
     client_patch = mock.patch(
-        "google.cloud.bigquery.magics.bigquery.Client", autospec=True
+        "google.cloud.bigquery.magics.magics.bigquery.Client", autospec=True
     )
 
     job_config = job.QueryJobConfig()
@@ -350,7 +358,7 @@ def test__create_dataset_if_necessary_exists():
     dataset_reference = bigquery.dataset.DatasetReference(project, dataset_id)
     dataset = bigquery.Dataset(dataset_reference)
     client_patch = mock.patch(
-        "google.cloud.bigquery.magics.bigquery.Client", autospec=True
+        "google.cloud.bigquery.magics.magics.bigquery.Client", autospec=True
     )
     with client_patch as client_mock:
         client = client_mock()
@@ -364,7 +372,7 @@ def test__create_dataset_if_necessary_not_exist():
     project = "project_id"
     dataset_id = "dataset_id"
     client_patch = mock.patch(
-        "google.cloud.bigquery.magics.bigquery.Client", autospec=True
+        "google.cloud.bigquery.magics.magics.bigquery.Client", autospec=True
     )
     with client_patch as client_mock:
         client = client_mock()
@@ -382,7 +390,7 @@ def test_extension_load():
 
     # verify that the magic is registered and has the correct source
     magic = ip.magics_manager.magics["cell"].get("bigquery")
-    assert magic.__module__ == "google.cloud.bigquery.magics"
+    assert magic.__module__ == "google.cloud.bigquery.magics.magics"
 
 
 @pytest.mark.usefixtures("ipython_interactive")
@@ -415,7 +423,7 @@ def test_bigquery_magic_without_optional_arguments(monkeypatch):
     sql = "SELECT 17 AS num"
     result = pandas.DataFrame([17], columns=["num"])
     run_query_patch = mock.patch(
-        "google.cloud.bigquery.magics._run_query", autospec=True
+        "google.cloud.bigquery.magics.magics._run_query", autospec=True
     )
     query_job_mock = mock.create_autospec(
         google.cloud.bigquery.job.QueryJob, instance=True
@@ -445,7 +453,7 @@ def test_bigquery_magic_default_connection_user_agent():
         "google.auth.default", return_value=(credentials_mock, "general-project")
     )
     run_query_patch = mock.patch(
-        "google.cloud.bigquery.magics._run_query", autospec=True
+        "google.cloud.bigquery.magics.magics._run_query", autospec=True
     )
     conn_patch = mock.patch("google.cloud.bigquery.client.Connection", autospec=True)
 
@@ -466,7 +474,7 @@ def test_bigquery_magic_with_legacy_sql():
     )
 
     run_query_patch = mock.patch(
-        "google.cloud.bigquery.magics._run_query", autospec=True
+        "google.cloud.bigquery.magics.magics._run_query", autospec=True
     )
     with run_query_patch as run_query_mock:
         ip.run_cell_magic("bigquery", "--use_legacy_sql", "SELECT 17 AS num")
@@ -477,19 +485,21 @@ def test_bigquery_magic_with_legacy_sql():
 
 @pytest.mark.usefixtures("ipython_interactive")
 @pytest.mark.skipif(pandas is None, reason="Requires `pandas`")
-def test_bigquery_magic_with_result_saved_to_variable():
+def test_bigquery_magic_with_result_saved_to_variable(ipython_ns_cleanup):
     ip = IPython.get_ipython()
     ip.extension_manager.load_extension("google.cloud.bigquery")
     magics.context.credentials = mock.create_autospec(
         google.auth.credentials.Credentials, instance=True
     )
 
+    ipython_ns_cleanup.append((ip, "df"))
+
     sql = "SELECT 17 AS num"
     result = pandas.DataFrame([17], columns=["num"])
     assert "df" not in ip.user_ns
 
     run_query_patch = mock.patch(
-        "google.cloud.bigquery.magics._run_query", autospec=True
+        "google.cloud.bigquery.magics.magics._run_query", autospec=True
     )
     query_job_mock = mock.create_autospec(
         google.cloud.bigquery.job.QueryJob, instance=True
@@ -516,10 +526,10 @@ def test_bigquery_magic_does_not_clear_display_in_verbose_mode():
     )
 
     clear_patch = mock.patch(
-        "google.cloud.bigquery.magics.display.clear_output", autospec=True
+        "google.cloud.bigquery.magics.magics.display.clear_output", autospec=True,
     )
     run_query_patch = mock.patch(
-        "google.cloud.bigquery.magics._run_query", autospec=True
+        "google.cloud.bigquery.magics.magics._run_query", autospec=True
     )
     with clear_patch as clear_mock, run_query_patch:
         ip.run_cell_magic("bigquery", "--verbose", "SELECT 17 as num")
@@ -536,10 +546,10 @@ def test_bigquery_magic_clears_display_in_verbose_mode():
     )
 
     clear_patch = mock.patch(
-        "google.cloud.bigquery.magics.display.clear_output", autospec=True
+        "google.cloud.bigquery.magics.magics.display.clear_output", autospec=True,
     )
     run_query_patch = mock.patch(
-        "google.cloud.bigquery.magics._run_query", autospec=True
+        "google.cloud.bigquery.magics.magics._run_query", autospec=True
     )
     with clear_patch as clear_mock, run_query_patch:
         ip.run_cell_magic("bigquery", "", "SELECT 17 as num")
@@ -576,7 +586,7 @@ def test_bigquery_magic_with_bqstorage_from_argument(monkeypatch):
     sql = "SELECT 17 AS num"
     result = pandas.DataFrame([17], columns=["num"])
     run_query_patch = mock.patch(
-        "google.cloud.bigquery.magics._run_query", autospec=True
+        "google.cloud.bigquery.magics.magics._run_query", autospec=True
     )
     query_job_mock = mock.create_autospec(
         google.cloud.bigquery.job.QueryJob, instance=True
@@ -635,7 +645,7 @@ def test_bigquery_magic_with_rest_client_requested(monkeypatch):
     sql = "SELECT 17 AS num"
     result = pandas.DataFrame([17], columns=["num"])
     run_query_patch = mock.patch(
-        "google.cloud.bigquery.magics._run_query", autospec=True
+        "google.cloud.bigquery.magics.magics._run_query", autospec=True
     )
     query_job_mock = mock.create_autospec(
         google.cloud.bigquery.job.QueryJob, instance=True
@@ -719,7 +729,7 @@ def test_bigquery_magic_w_max_results_query_job_results_fails():
         "google.cloud.bigquery.client.Client.query", autospec=True
     )
     close_transports_patch = mock.patch(
-        "google.cloud.bigquery.magics._close_transports", autospec=True,
+        "google.cloud.bigquery.magics.magics._close_transports", autospec=True,
     )
 
     sql = "SELECT 17 AS num"
@@ -751,7 +761,7 @@ def test_bigquery_magic_w_table_id_invalid():
     )
 
     list_rows_patch = mock.patch(
-        "google.cloud.bigquery.magics.bigquery.Client.list_rows",
+        "google.cloud.bigquery.magics.magics.bigquery.Client.list_rows",
         autospec=True,
         side_effect=exceptions.BadRequest("Not a valid table ID"),
     )
@@ -792,11 +802,13 @@ def test_bigquery_magic_w_missing_query():
 
 @pytest.mark.usefixtures("ipython_interactive")
 @pytest.mark.skipif(pandas is None, reason="Requires `pandas`")
-def test_bigquery_magic_w_table_id_and_destination_var():
+def test_bigquery_magic_w_table_id_and_destination_var(ipython_ns_cleanup):
     ip = IPython.get_ipython()
     ip.extension_manager.load_extension("google.cloud.bigquery")
     magics.context._project = None
 
+    ipython_ns_cleanup.append((ip, "df"))
+
     credentials_mock = mock.create_autospec(
         google.auth.credentials.Credentials, instance=True
     )
@@ -809,7 +821,7 @@ def test_bigquery_magic_w_table_id_and_destination_var():
     )
 
     client_patch = mock.patch(
-        "google.cloud.bigquery.magics.bigquery.Client", autospec=True
+        "google.cloud.bigquery.magics.magics.bigquery.Client", autospec=True
     )
 
     table_id = "bigquery-public-data.samples.shakespeare"
@@ -849,7 +861,7 @@ def test_bigquery_magic_w_table_id_and_bqstorage_client():
     )
 
     client_patch = mock.patch(
-        "google.cloud.bigquery.magics.bigquery.Client", autospec=True
+        "google.cloud.bigquery.magics.magics.bigquery.Client", autospec=True
     )
 
     bqstorage_mock = mock.create_autospec(bigquery_storage_v1.BigQueryReadClient)
@@ -882,7 +894,7 @@ def test_bigquery_magic_dryrun_option_sets_job_config():
     )
 
     run_query_patch = mock.patch(
-        "google.cloud.bigquery.magics._run_query", autospec=True
+        "google.cloud.bigquery.magics.magics._run_query", autospec=True
     )
 
     sql = "SELECT 17 AS num"
@@ -905,7 +917,7 @@ def test_bigquery_magic_dryrun_option_returns_query_job():
         google.cloud.bigquery.job.QueryJob, instance=True
     )
     run_query_patch = mock.patch(
-        "google.cloud.bigquery.magics._run_query", autospec=True
+        "google.cloud.bigquery.magics.magics._run_query", autospec=True
     )
 
     sql = "SELECT 17 AS num"
@@ -919,15 +931,17 @@ def test_bigquery_magic_dryrun_option_returns_query_job():
 
 
 @pytest.mark.usefixtures("ipython_interactive")
-def test_bigquery_magic_dryrun_option_variable_error_message():
+def test_bigquery_magic_dryrun_option_variable_error_message(ipython_ns_cleanup):
     ip = IPython.get_ipython()
     ip.extension_manager.load_extension("google.cloud.bigquery")
     magics.context.credentials = mock.create_autospec(
         google.auth.credentials.Credentials, instance=True
     )
 
+    ipython_ns_cleanup.append((ip, "q_job"))
+
     run_query_patch = mock.patch(
-        "google.cloud.bigquery.magics._run_query",
+        "google.cloud.bigquery.magics.magics._run_query",
         autospec=True,
         side_effect=exceptions.BadRequest("Syntax error in SQL query"),
     )
@@ -944,7 +958,7 @@ def test_bigquery_magic_dryrun_option_variable_error_message():
 
 
 @pytest.mark.usefixtures("ipython_interactive")
-def test_bigquery_magic_dryrun_option_saves_query_job_to_variable():
+def test_bigquery_magic_dryrun_option_saves_query_job_to_variable(ipython_ns_cleanup):
     ip = IPython.get_ipython()
     ip.extension_manager.load_extension("google.cloud.bigquery")
     magics.context.credentials = mock.create_autospec(
@@ -954,9 +968,11 @@ def test_bigquery_magic_dryrun_option_saves_query_job_to_variable():
         google.cloud.bigquery.job.QueryJob, instance=True
     )
     run_query_patch = mock.patch(
-        "google.cloud.bigquery.magics._run_query", autospec=True
+        "google.cloud.bigquery.magics.magics._run_query", autospec=True
     )
 
+    ipython_ns_cleanup.append((ip, "q_job"))
+
     sql = "SELECT 17 AS num"
 
     assert "q_job" not in ip.user_ns
@@ -972,13 +988,15 @@ def test_bigquery_magic_dryrun_option_saves_query_job_to_variable():
 
 
 @pytest.mark.usefixtures("ipython_interactive")
-def test_bigquery_magic_saves_query_job_to_variable_on_error():
+def test_bigquery_magic_saves_query_job_to_variable_on_error(ipython_ns_cleanup):
     ip = IPython.get_ipython()
     ip.extension_manager.load_extension("google.cloud.bigquery")
     magics.context.credentials = mock.create_autospec(
         google.auth.credentials.Credentials, instance=True
     )
 
+    ipython_ns_cleanup.append((ip, "result"))
+
     client_query_patch = mock.patch(
         "google.cloud.bigquery.client.Client.query", autospec=True
     )
@@ -1151,7 +1169,7 @@ def test_bigquery_magic_with_project():
         "google.auth.default", return_value=(credentials_mock, "general-project")
     )
     run_query_patch = mock.patch(
-        "google.cloud.bigquery.magics._run_query", autospec=True
+        "google.cloud.bigquery.magics.magics._run_query", autospec=True
     )
     with run_query_patch as run_query_mock, default_patch:
         ip.run_cell_magic("bigquery", "--project=specific-project", "SELECT 17 as num")
@@ -1162,30 +1180,65 @@ def test_bigquery_magic_with_project():
         assert magics.context.project == "general-project"
 
 
+@pytest.mark.usefixtures("ipython_interactive")
+def test_bigquery_magic_with_multiple_options():
+    ip = IPython.get_ipython()
+    ip.extension_manager.load_extension("google.cloud.bigquery")
+    magics.context._project = None
+
+    credentials_mock = mock.create_autospec(
+        google.auth.credentials.Credentials, instance=True
+    )
+    default_patch = mock.patch(
+        "google.auth.default", return_value=(credentials_mock, "general-project")
+    )
+    run_query_patch = mock.patch(
+        "google.cloud.bigquery.magics.magics._run_query", autospec=True
+    )
+    with run_query_patch as run_query_mock, default_patch:
+        ip.run_cell_magic(
+            "bigquery",
+            "--project=specific-project --use_legacy_sql --maximum_bytes_billed 1024",
+            "SELECT 17 as num",
+        )
+
+    args, kwargs = run_query_mock.call_args
+    client_used = args[0]
+    assert client_used.project == "specific-project"
+
+    job_config_used = kwargs["job_config"]
+    assert job_config_used.use_legacy_sql
+    assert job_config_used.maximum_bytes_billed == 1024
+
+
 @pytest.mark.usefixtures("ipython_interactive")
 @pytest.mark.skipif(pandas is None, reason="Requires `pandas`")
-def test_bigquery_magic_with_string_params():
+def test_bigquery_magic_with_string_params(ipython_ns_cleanup):
     ip = IPython.get_ipython()
     ip.extension_manager.load_extension("google.cloud.bigquery")
     magics.context.credentials = mock.create_autospec(
         google.auth.credentials.Credentials, instance=True
     )
 
+    ipython_ns_cleanup.append((ip, "params_dict_df"))
+
     sql = "SELECT @num AS num"
     result = pandas.DataFrame([17], columns=["num"])
-    assert "params_string_df" not in ip.user_ns
+
+    assert "params_dict_df" not in ip.user_ns
 
     run_query_patch = mock.patch(
-        "google.cloud.bigquery.magics._run_query", autospec=True
+        "google.cloud.bigquery.magics.magics._run_query", autospec=True
     )
     query_job_mock = mock.create_autospec(
         google.cloud.bigquery.job.QueryJob, instance=True
     )
     query_job_mock.to_dataframe.return_value = result
+
     with run_query_patch as run_query_mock:
         run_query_mock.return_value = query_job_mock
 
-        ip.run_cell_magic("bigquery", 'params_string_df --params {"num":17}', sql)
+        ip.run_cell_magic("bigquery", "params_string_df --params='{\"num\":17}'", sql)
 
         run_query_mock.assert_called_once_with(mock.ANY, sql.format(num=17), mock.ANY)
 
@@ -1197,19 +1250,24 @@ def test_bigquery_magic_with_string_params():
 
 @pytest.mark.usefixtures("ipython_interactive")
 @pytest.mark.skipif(pandas is None, reason="Requires `pandas`")
-def test_bigquery_magic_with_dict_params():
+def test_bigquery_magic_with_dict_params(ipython_ns_cleanup):
     ip = IPython.get_ipython()
     ip.extension_manager.load_extension("google.cloud.bigquery")
     magics.context.credentials = mock.create_autospec(
         google.auth.credentials.Credentials, instance=True
     )
 
-    sql = "SELECT @num AS num"
-    result = pandas.DataFrame([17], columns=["num"])
+    ipython_ns_cleanup.append((ip, "params_dict_df"))
+
+    sql = "SELECT @num AS num, @tricky_value as tricky_value"
+    result = pandas.DataFrame(
+        [(False, '--params "value"')], columns=["valid", "tricky_value"]
+    )
+
     assert "params_dict_df" not in ip.user_ns
 
     run_query_patch = mock.patch(
-        "google.cloud.bigquery.magics._run_query", autospec=True
+        "google.cloud.bigquery.magics.magics._run_query", autospec=True
     )
     query_job_mock = mock.create_autospec(
         google.cloud.bigquery.job.QueryJob, instance=True
@@ -1218,7 +1276,7 @@ def test_bigquery_magic_with_dict_params():
     with run_query_patch as run_query_mock:
         run_query_mock.return_value = query_job_mock
 
-        params = {"num": 17}
+        params = {"valid": False, "tricky_value": '--params "value"'}
         # Insert dictionary into user namespace so that it can be expanded
         ip.user_ns["params"] = params
         ip.run_cell_magic("bigquery", "params_dict_df --params $params", sql)
@@ -1230,6 +1288,194 @@ def test_bigquery_magic_with_dict_params():
     assert len(df) == len(result)  # verify row count
     assert list(df) == list(result)  # verify column names
 
+    assert not df["valid"][0]
+    assert df["tricky_value"][0] == '--params "value"'
+
+
+@pytest.mark.usefixtures("ipython_interactive")
+@pytest.mark.skipif(pandas is None, reason="Requires `pandas`")
+def test_bigquery_magic_with_dict_params_nonexisting():
+    ip = IPython.get_ipython()
+    ip.extension_manager.load_extension("google.cloud.bigquery")
+    magics.context.credentials = mock.create_autospec(
+        google.auth.credentials.Credentials, instance=True
+    )
+
+    sql = "SELECT @foo AS foo"
+
+    with pytest.raises(NameError, match=r".*undefined variable.*unknown_name.*"):
+        ip.run_cell_magic("bigquery", "params_dict_df --params $unknown_name", sql)
+
+
+@pytest.mark.usefixtures("ipython_interactive")
+@pytest.mark.skipif(pandas is None, reason="Requires `pandas`")
+def test_bigquery_magic_with_dict_params_incorrect_syntax():
+    ip = IPython.get_ipython()
+    ip.extension_manager.load_extension("google.cloud.bigquery")
+    magics.context.credentials = mock.create_autospec(
+        google.auth.credentials.Credentials, instance=True
+    )
+
+    sql = "SELECT @foo AS foo"
+
+    with pytest.raises(SyntaxError, match=r".*--params.*"):
+        cell_magic_args = "params_dict_df --params {'foo': 1; 'bar': 2}"
+        ip.run_cell_magic("bigquery", cell_magic_args, sql)
+
+
+@pytest.mark.usefixtures("ipython_interactive")
+@pytest.mark.skipif(pandas is None, reason="Requires `pandas`")
+def test_bigquery_magic_with_dict_params_duplicate():
+    ip = IPython.get_ipython()
+    ip.extension_manager.load_extension("google.cloud.bigquery")
+    magics.context.credentials = mock.create_autospec(
+        google.auth.credentials.Credentials, instance=True
+    )
+
+    sql = "SELECT @foo AS foo"
+
+    with pytest.raises(ValueError, match=r"Duplicate --params option\."):
+        cell_magic_args = (
+            "params_dict_df --params {'foo': 1} --verbose --params {'bar': 2} "
+        )
+        ip.run_cell_magic("bigquery", cell_magic_args, sql)
+
+
+@pytest.mark.usefixtures("ipython_interactive")
+@pytest.mark.skipif(pandas is None, reason="Requires `pandas`")
+def test_bigquery_magic_with_option_value_incorrect():
+    ip = IPython.get_ipython()
+    ip.extension_manager.load_extension("google.cloud.bigquery")
+    magics.context.credentials = mock.create_autospec(
+        google.auth.credentials.Credentials, instance=True
+    )
+
+    sql = "SELECT @foo AS foo"
+
+    with pytest.raises(ValueError, match=r".*invalid literal.*\[PLENTY!\].*"):
+        cell_magic_args = "params_dict_df --max_results [PLENTY!]"
+        ip.run_cell_magic("bigquery", cell_magic_args, sql)
+
+
+@pytest.mark.usefixtures("ipython_interactive")
+@pytest.mark.skipif(pandas is None, reason="Requires `pandas`")
+def test_bigquery_magic_with_dict_params_negative_value(ipython_ns_cleanup):
+    ip = IPython.get_ipython()
+    ip.extension_manager.load_extension("google.cloud.bigquery")
+    magics.context.credentials = mock.create_autospec(
+        google.auth.credentials.Credentials, instance=True
+    )
+
+    ipython_ns_cleanup.append((ip, "params_dict_df"))
+
+    sql = "SELECT @num AS num"
+    result = pandas.DataFrame([-17], columns=["num"])
+
+    assert "params_dict_df" not in ip.user_ns
+
+    run_query_patch = mock.patch(
+        "google.cloud.bigquery.magics.magics._run_query", autospec=True
+    )
+    query_job_mock = mock.create_autospec(
+        google.cloud.bigquery.job.QueryJob, instance=True
+    )
+    query_job_mock.to_dataframe.return_value = result
+    with run_query_patch as run_query_mock:
+        run_query_mock.return_value = query_job_mock
+
+        params = {"num": -17}
+        # Insert dictionary into user namespace so that it can be expanded
+        ip.user_ns["params"] = params
+        ip.run_cell_magic("bigquery", "params_dict_df --params $params", sql)
+
+        run_query_mock.assert_called_once_with(mock.ANY, sql.format(num=-17), mock.ANY)
+
+    assert "params_dict_df" in ip.user_ns  # verify that the variable exists
+    df = ip.user_ns["params_dict_df"]
+    assert len(df) == len(result)  # verify row count
+    assert list(df) == list(result)  # verify column names
+    assert df["num"][0] == -17
+
+
+@pytest.mark.usefixtures("ipython_interactive")
+@pytest.mark.skipif(pandas is None, reason="Requires `pandas`")
+def test_bigquery_magic_with_dict_params_array_value(ipython_ns_cleanup):
+    ip = IPython.get_ipython()
+    ip.extension_manager.load_extension("google.cloud.bigquery")
+    magics.context.credentials = mock.create_autospec(
+        google.auth.credentials.Credentials, instance=True
+    )
+
+    ipython_ns_cleanup.append((ip, "params_dict_df"))
+
+    sql = "SELECT @num AS num"
+    result = pandas.DataFrame(["foo bar", "baz quux"], columns=["array_data"])
+
+    assert "params_dict_df" not in ip.user_ns
+
+    run_query_patch = mock.patch(
+        "google.cloud.bigquery.magics.magics._run_query", autospec=True
+    )
+    query_job_mock = mock.create_autospec(
+        google.cloud.bigquery.job.QueryJob, instance=True
+    )
+    query_job_mock.to_dataframe.return_value = result
+    with run_query_patch as run_query_mock:
+        run_query_mock.return_value = query_job_mock
+
+        params = {"array_data": ["foo bar", "baz quux"]}
+        # Insert dictionary into user namespace so that it can be expanded
+        ip.user_ns["params"] = params
+        ip.run_cell_magic("bigquery", "params_dict_df --params $params", sql)
+
+        run_query_mock.assert_called_once_with(mock.ANY, sql.format(num=-17), mock.ANY)
+
+    assert "params_dict_df" in ip.user_ns  # verify that the variable exists
+    df = ip.user_ns["params_dict_df"]
+    assert len(df) == len(result)  # verify row count
+    assert list(df) == list(result)  # verify column names
+    assert list(df["array_data"]) == ["foo bar", "baz quux"]
+
+
+@pytest.mark.usefixtures("ipython_interactive")
+@pytest.mark.skipif(pandas is None, reason="Requires `pandas`")
+def test_bigquery_magic_with_dict_params_tuple_value(ipython_ns_cleanup):
+    ip = IPython.get_ipython()
+    ip.extension_manager.load_extension("google.cloud.bigquery")
+    magics.context.credentials = mock.create_autospec(
+        google.auth.credentials.Credentials, instance=True
+    )
+
+    ipython_ns_cleanup.append((ip, "params_dict_df"))
+
+    sql = "SELECT @num AS num"
+    result = pandas.DataFrame(["foo bar", "baz quux"], columns=["array_data"])
+
+    assert "params_dict_df" not in ip.user_ns
+
+    run_query_patch = mock.patch(
+        "google.cloud.bigquery.magics.magics._run_query", autospec=True
+    )
+    query_job_mock = mock.create_autospec(
+        google.cloud.bigquery.job.QueryJob, instance=True
+    )
+    query_job_mock.to_dataframe.return_value = result
+    with run_query_patch as run_query_mock:
+        run_query_mock.return_value = query_job_mock
+
+        params = {"array_data": ("foo bar", "baz quux")}
+        # Insert dictionary into user namespace so that it can be expanded
+        ip.user_ns["params"] = params
+        ip.run_cell_magic("bigquery", "params_dict_df --params $params", sql)
+
+        run_query_mock.assert_called_once_with(mock.ANY, sql.format(num=-17), mock.ANY)
+
+    assert "params_dict_df" in ip.user_ns  # verify that the variable exists
+    df = ip.user_ns["params_dict_df"]
+    assert len(df) == len(result)  # verify row count
+    assert list(df) == list(result)  # verify column names
+    assert list(df["array_data"]) == ["foo bar", "baz quux"]
+
 
 @pytest.mark.usefixtures("ipython_interactive")
 @pytest.mark.skipif(pandas is None, reason="Requires `pandas`")
@@ -1246,6 +1492,24 @@ def test_bigquery_magic_with_improperly_formatted_params():
         ip.run_cell_magic("bigquery", "--params {17}", sql)
 
 
+@pytest.mark.usefixtures("ipython_interactive")
+@pytest.mark.skipif(pandas is None, reason="Requires `pandas`")
+def test_bigquery_magic_with_invalid_multiple_option_values():
+    ip = IPython.get_ipython()
+    ip.extension_manager.load_extension("google.cloud.bigquery")
+    magics.context.credentials = mock.create_autospec(
+        google.auth.credentials.Credentials, instance=True
+    )
+
+    sql = "SELECT @foo AS foo"
+
+    exc_pattern = r".*[Uu]nrecognized input.*option values correct\?.*567.*"
+
+    with pytest.raises(ValueError, match=exc_pattern):
+        cell_magic_args = "params_dict_df --max_results 10 567"
+        ip.run_cell_magic("bigquery", cell_magic_args, sql)
+
+
 @pytest.mark.usefixtures("ipython_interactive")
 def test_bigquery_magic_omits_tracebacks_from_error_message():
     ip = IPython.get_ipython()
@@ -1259,7 +1523,7 @@ def test_bigquery_magic_omits_tracebacks_from_error_message():
     )
 
     run_query_patch = mock.patch(
-        "google.cloud.bigquery.magics._run_query",
+        "google.cloud.bigquery.magics.magics._run_query",
         autospec=True,
         side_effect=exceptions.BadRequest("Syntax error in SQL query"),
     )
@@ -1287,7 +1551,7 @@ def test_bigquery_magic_w_destination_table_invalid_format():
     )
 
     client_patch = mock.patch(
-        "google.cloud.bigquery.magics.bigquery.Client", autospec=True
+        "google.cloud.bigquery.magics.magics.bigquery.Client", autospec=True
     )
 
     with client_patch, default_patch, pytest.raises(ValueError) as exc_context:
@@ -1310,11 +1574,12 @@ def test_bigquery_magic_w_destination_table():
     )
 
     create_dataset_if_necessary_patch = mock.patch(
-        "google.cloud.bigquery.magics._create_dataset_if_necessary", autospec=True
+        "google.cloud.bigquery.magics.magics._create_dataset_if_necessary",
+        autospec=True,
     )
 
     run_query_patch = mock.patch(
-        "google.cloud.bigquery.magics._run_query", autospec=True
+        "google.cloud.bigquery.magics.magics._run_query", autospec=True
     )
 
     with create_dataset_if_necessary_patch, run_query_patch as run_query_mock:
@@ -1341,12 +1606,12 @@ def test_bigquery_magic_create_dataset_fails():
     )
 
     create_dataset_if_necessary_patch = mock.patch(
-        "google.cloud.bigquery.magics._create_dataset_if_necessary",
+        "google.cloud.bigquery.magics.magics._create_dataset_if_necessary",
         autospec=True,
         side_effect=OSError,
     )
     close_transports_patch = mock.patch(
-        "google.cloud.bigquery.magics._close_transports", autospec=True,
+        "google.cloud.bigquery.magics.magics._close_transports", autospec=True,
     )
 
     with pytest.raises(
